# -*- coding: utf-8 -*-
import sqlalchemy
import sqlalchemy.orm as orm
from ..version import __version__
from zope.sqlalchemy import ZopeTransactionExtension

DBSession = None


def pkg_prefix(config):
    '''
    Function for return pkg prefix.

    >>> from pyramid.config import Configurator
    >>> settings = {'foo': 'foo', 'bar': 'bar'}

    # Create config
    >>> config = Configurator(settings=settings)

    # w/o route_prefix
    >>> pkg_prefix(config)
    '/sacrud/'

    # with route_prefix
    >>> config.route_prefix = "/admin"
    >>> pkg_prefix(config)
    ''
    '''
    return '' if config.route_prefix else '/sacrud/'


def add_routes(config):
    prefix = pkg_prefix(config)
    config.add_route('sa_home',     prefix)
    config.add_route('sa_list',     prefix + '{table}')
    config.add_route('sa_create',   prefix + '{table}/create')
    config.add_route('sa_read',     prefix + '{table}/read/{id}')
    config.add_route('sa_update',   prefix + '{table}/update/{id}')
    config.add_route('sa_delete',   prefix + '{table}/delete/{id}')
    config.add_route('sa_paste',    prefix + '{table}/paste/{id}/' +
                                             '{target_id}')
    config.add_route('sa_paste_tmp', prefix + '{table}/paste/{id}')
    config.add_route('sa_union_fields', prefix + '{table}/union')


def includeme(config):
    global DBSession
    engine = sqlalchemy.engine_from_config(config.registry.settings)
    if DBSession is None:
        DBSession = orm.scoped_session(
            orm.sessionmaker(extension=ZopeTransactionExtension()))
    DBSession.remove()
    DBSession.configure(bind=engine)

    config.include('pyramid_jinja2')
    config.add_static_view('/sa_static', 'sacrud:static')
    config.include(add_routes)
    config.add_jinja2_search_path("sacrud:templates")
    env = config.get_jinja2_environment()

    # if variable is None print '' instead of 'None'
    def _silent_none(value):
        if value is None:
            return ''
        return value
    env.finalize = _silent_none
    config.add_jinja2_extension('jinja2.ext.do')
    env.globals['str'] = str
    env.globals['getattr'] = getattr
    env.globals['isinstance'] = isinstance
    env.globals['sqlalchemy'] = sqlalchemy
    env.globals['session'] = DBSession
    env.globals['sacrud_ver'] = __version__
    config.scan()



def before_insert(mapper, connection, target):

    cls = target.__class__
    pos_name = target.__mapper_args__['order_by']
    pos_field = target.__getattribute__(pos_name)
    position = pos_field and pos_field or 0

    connection.execute(cls.__table__.
                       update(values={pos_name: getattr(cls, pos_name) + 1}).
                       where(cls.id != target.id).
                       where(getattr(cls, pos_name) >= position)
                       )



# -*- coding: utf-8 -*-
import ast
import sqlalchemy as sa
from sacrud import action
from pyramid.httpexceptions import HTTPFound
from pyramid.view import view_config
from sacrud.pyramid_ext import DBSession


def breadcrumbs(tname,  view, id=None):
    bc = {}
    bc['sa_list'] = [{'name': 'Home', 'visible': True,
                      'view': 'sa_home',
                      'param': {'table': tname}},
                     {'name': tname, 'visible': True,
                      'view': 'sa_list',
                      'param': {'table': tname}}]
    bc['sa_create'] = bc['sa_list'][:]
    bc['sa_create'].append({'name': 'create',
                            'visible': False,
                            'view': 'sa_list',
                            'param': {'table': tname}})
    bc['sa_read'] = bc['sa_list'][:]
    bc['sa_read'].append({'name': id,
                          'visible': False,
                          'view': 'sa_list',
                          'param': {'table': tname}})
    bc['sa_update'] = bc['sa_read']
    bc['sa_union'] = bc['sa_list'][:]
    bc['sa_union'].append({'name': 'union',
                           'visible': False,
                           'view': 'sa_list',
                           'param': {'table': tname}})

    return bc[view]


def get_models(request):
    settings = request.registry.settings
    return settings['sacrud_models']


def get_table(tname, request):
    tables = get_models(request)
    return filter(lambda table: (table.__tablename__).
                  lower() == tname.lower(), tables)[0]


def get_relationship(tname, request):
    obj = get_table(tname, request)
    # Build a list of only relationship properties
    relation_properties = filter(
        lambda p: isinstance(p, sa.orm.properties.RelationshipProperty),
        sa.orm.class_mapper(obj).iterate_properties
    )
    #related_tables = [prop.target for prop in relation_properties]
    related_classes = [{'cls': prop.mapper.class_,
                        'col': list(prop.local_columns)[0]}
                       for prop in relation_properties]
    return related_classes


@view_config(route_name='sa_home', renderer='/sacrud/home.jinja2')
def sa_home(request):
    return {'tables': get_models(request)}


@view_config(route_name='sa_list', renderer='/sacrud/list.jinja2')
def sa_list(request):
    tname = request.matchdict['table']
    order_by = request.params.get('order_by', False)
    args = [DBSession, get_table(tname, request)]
    if order_by:
        args.append(order_by)
    resp = action.index(*args)
    return {'sa_crud': resp, 'breadcrumbs': breadcrumbs(tname, 'sa_list')}


@view_config(route_name='sa_create', renderer='/sacrud/create.jinja2')
def sa_create(request):
    tname = request.matchdict['table']
    if 'form.submitted' in request.params:
        action.create(DBSession, get_table(tname, request),
                      request.params.dict_of_lists())
        return HTTPFound(location=request.route_url('sa_list', table=tname))
    resp = action.create(DBSession, get_table(tname, request))
    rel = get_relationship(tname, request)
    return {'sa_crud': resp, 'rel': rel,
            'breadcrumbs': breadcrumbs(tname, 'sa_create')}


@view_config(route_name='sa_read', renderer='/sacrud/read.jinja2')
def sa_read(request):
    tname = request.matchdict['table']
    id = request.matchdict['id']
    resp = action.read(DBSession, get_table(tname, request), id)
    return {'sa_crud': resp,
            'breadcrumbs': breadcrumbs(tname, 'sa_read', id=id)}


@view_config(route_name='sa_update', renderer='/sacrud/create.jinja2')
def sa_update(request):
    tname = request.matchdict['table']
    id = request.matchdict['id']
    if 'form.submitted' in request.params:
        action.update(DBSession, get_table(tname, request), id,
                      request.params.dict_of_lists())
        return HTTPFound(location=request.route_url('sa_list', table=tname))
    resp = action.update(DBSession, get_table(tname, request), id)
    rel = get_relationship(tname, request)
    return {'sa_crud': resp, 'rel': rel,
            'breadcrumbs': breadcrumbs(tname, 'sa_update', id=id)}


@view_config(route_name='sa_paste', renderer='/sacrud/list.jinja2')
def sa_paste(request):
    tname = request.matchdict['table']
    id = request.matchdict['id']
    target_id = request.matchdict['target_id']

    source_obj = action.read(DBSession, get_table(tname, request), id)['obj']
    pos_name = source_obj.__mapper_args__['order_by']
    action.update(DBSession, get_table(tname, request), target_id,
                  {pos_name: [getattr(source_obj, pos_name)]})

    return HTTPFound(location=request.route_url('sa_list', table=tname))


@view_config(route_name='sa_delete')
def sa_delete(request):
    tname = request.matchdict['table']
    id = request.matchdict['id']
    action.delete(DBSession, get_table(tname, request), id)
    return HTTPFound(location=request.route_url('sa_list', table=tname))


@view_config(route_name='sa_union_fields', renderer='/sacrud/compare.jinja2')
def sa_union_fields(request):
    tname = request.matchdict['table']
    table = get_table(tname, request)
    pk = action.get_pk(table)
    if 'form.submitted' in request.params and pk:
        comp_objs = ast.literal_eval(request.POST['sa_comp_objs'])
        new_row_param = {}
        for field in request.POST:
            if 'sa_col_' not in field:
                continue
            row_name = field.replace("sa_col_", "")
            row_id_for_value = request.POST[field]
            row_value = DBSession.query(getattr(table, row_name))\
                                 .filter(
                                     getattr(table, pk) == row_id_for_value
                                 ).one()
            new_row_param[row_name] = [row_value, ]
        # delete rows
        for id in comp_objs:
            action.delete(DBSession, table, id)
        # add new union row
        action.create(DBSession, table, new_row_param)
        return HTTPFound(location=request.route_url('sa_list', table=tname))

    order_by = request.params.get('order_by', False)
    args = [DBSession, table]
    if order_by:
        args.append(order_by)
    resp = action.index(*args)

    sa_checked_row = request.POST.get('sa_checked_row', None).split(",")
    if '' not in sa_checked_row:
        compare_objs = map(int, sa_checked_row)
    else:
        return HTTPFound(location=request.route_url('sa_list', table=tname))
    return {'compare_objs': compare_objs, 'sa_crud': resp,
            'breadcrumbs': breadcrumbs(tname, 'sa_union')}



# -*- coding: utf-8 -*-
import os
import ast
import uuid
import inspect
import sqlalchemy
import transaction

prefix = 'crud'


def get_pk(table):
    # FIXME: кривое исключение
    """ Return primary key name of table.

    :Parameters:
        - `table`: SQLAlchemy table.

    :Examples:

    >>> from sqlalchemy import Column, Integer
    >>> from sqlalchemy.ext.declarative import declarative_base
    >>> Base = declarative_base()
    >>> class User(Base):
    ...     __tablename__ = 'users'
    ...     id = Column(Integer, primary_key=True)
    >>> get_pk(User())
    'id'

    """
    try:
        pk = table.__mapper__.primary_key[0].name
    except:
        pk = ''

    return pk


def get_relations(obj):
    """
    :Examples:

    >>> from sqlalchemy import Column, Integer, ForeignKey
    >>> from sqlalchemy.ext.declarative import declarative_base
    >>> from sqlalchemy.orm import relationship, backref
    >>> Base = declarative_base()
    >>> class User(Base):
    ...    __tablename__ = 'users'
    ...    id = Column(Integer, primary_key=True)
    ...
    >>> class Address(Base):
    ...    __tablename__ = 'addresses'
    ...    id = Column(Integer, primary_key=True)
    ...    user_id = Column(Integer, ForeignKey('users.id'))
    ...    user = relationship("User", backref=backref('addresses', order_by=id))
    >>> get_relations(User())
    [('addresses', [])]

    """
    return [(n, getattr(obj, n)) for n in dir(obj)
            if isinstance(getattr(obj, n),
                          sqlalchemy.orm.collections.InstrumentedList)]


def index(session, table, order_by=None):
    """
    Return a list of table rows.

    :Parameters:

        - `session`: DBSession.
        - `table`: table instance.
        - `order_by`: name ordered row.
    """
    col = [c for c in table.__table__.columns]
    pk_name = get_pk(table)
    query = session.query(table)
    if order_by:
        query = query.order_by(order_by)
    row = query.all()
    if hasattr(table, '__mapper_args__'):
        mapper_args = table.__mapper_args__
    else:
        mapper_args = {}

    return {'row': row,
            'pk': pk_name,
            'col': col,
            'table': table,
            'prefix': prefix,
            'mapper_args': mapper_args, }


def create(session, table, request=''):
    """
    Insert row to table.

    :Parameters:

        - `session`: DBSession.
        - `table`: table instance.
        - `request`: webob format request.
    """
    if request:
        args = {}
        # FIXME: я чувствую здесь диссонанс
        for arg in inspect.getargspec(table.__init__).args[1:]:
            args[arg] = None
        obj = table(**args)
        for key, value in request.items():
            # chek columns exist
            if not key in table.__table__.columns:
                continue
            value = check_type(request, table, key)
            obj.__setattr__(key, value)
        session.add(obj)
        transaction.commit()
        return

    pk_name = get_pk(table)
    col = [c for c in table.__table__.columns]
    return {'pk': pk_name,
            'col': col,
            'table': table,
            'prefix': prefix}


def read(session, table, pk):
    """
    Select row by pk.

    :Parameters:

        - `session`: DBSession.
        - `table`: table instance.
        - `pk`: primary key value.
    """
    pk_name = get_pk(table)
    obj = session.query(table).filter(getattr(table, pk_name) == pk).one()
    col = [c for c in table.__table__.columns]
    return {'obj': obj,
            'pk': pk_name,
            'col': col,
            'table': table,
            'prefix': prefix,
            'rel': get_relations(obj)}


def update(session, table, pk, request=''):
    """
    Update row of table.

    :Parameters:

        - `session`: DBSession.
        - `table`: table instance.
        - `request`: webob format request.
    """

    pk_name = get_pk(table)
    obj = session.query(table).filter(getattr(table, pk_name) == pk).one()
    col_list = [c for c in table.__table__.columns]

    if request:
        for col in col_list:
            if col.name not in request:
                continue
            if getattr(obj, col.name) == request[col.name][0]:
                continue
            if col.type.__class__.__name__ == 'FileStore':
                if not hasattr(request[col.name][0], 'filename'):
                    continue
            value = check_type(request, table, col.name, obj)
            setattr(obj, col.name, value)
        session.add(obj)
        transaction.commit()
        return

    return {'obj': obj,
            'pk': pk_name,
            'col': col_list,
            'table': table,
            'prefix': prefix}


def delete(session, table, pk):
    """
    Delete row by pk.

    :Parameters:

        - `session`: DBSession.
        - `table`: table instance.
        - `pk`: primary key value.
    """

    pk_name = get_pk(table)
    obj = session.query(table).filter(getattr(table, pk_name) == pk).one()
    check_type('', table, obj=obj)
    session.delete(obj)
    transaction.commit()


def delete_fileobj(table, obj, key):
    """ Delete atached file.
    """
    abspath = table.__table__.columns[key].type.abspath
    path = os.path.join(abspath, os.path.basename(getattr(obj, key)))
    if not obj or not os.path.isfile(path):
        return
    os.remove(path)


def check_type(request, table, key=None, obj=None):
    """ Chek type when Create, Update or Delete.
    """

    # for Delete
    if not key:
        for col in table.__table__.columns:
            if col.type.__class__.__name__ == 'FileStore':
                if getattr(obj, col.name):
                    delete_fileobj(table, obj, col.name)
        return
    column_type = table.__table__.columns[key].type.__class__.__name__

    # for Update or Create
    value = request[key]
    if type(value) in (list, tuple):
        value = value[0]
    if column_type == 'Boolean':
        value = False if value == '0' else True
        value = True if value else False
    elif column_type == 'FileStore':
        if request[key] is None:
            return None
        fileobj = request[key][0]
        if hasattr(fileobj, 'filename'):
            extension = fileobj.filename.split(".")[-1]
            fileobj.filename = str(uuid.uuid4()) + "." + extension
            abspath = table.__table__.columns[key].type.abspath
            store_file(request, key, abspath)
            if obj:
                if getattr(obj, key):
                    delete_fileobj(table, obj, key)
            value = fileobj.filename
    elif column_type == 'HSTORE':
        value = ast.literal_eval(value)
    return value


def store_file(request, key, path):
    """ Load atached file.
    """
    # ``filename`` contains the name of the file in string format.
    #
    # WARNING: this example does not deal with the fact that IE sends an
    # absolute file *path* as the filename.  This example is naive; it
    # trusts user input.
    filename = request[key][0].filename

    # ``input_file`` contains the actual file data which needs to be
    # stored somewhere.

    input_file = request[key][0].file

    # Using the filename like this without cleaning it is very
    # insecure so please keep that in mind when writing your own
    # file handlingself.
    file_path = os.path.join(path, filename)
    output_file = open(file_path, 'wb')

    # Finally write the data to the output file
    input_file.seek(0)
    while 1:
        data = input_file.read(2 << 16)
        if not data:
            break
        output_file.write(data)
    output_file.close()



import os
import glob
import fnmatch
from sacrud import version
from setuptools import setup


def opj(*args):
    path = os.path.join(*args)
    return os.path.normpath(path)

badnames = [".pyc", ".py", "~", "no_"]


def find_data_files(srcdir, *wildcards, **kw):
    # get a list of all files under the srcdir matching wildcards,
    # returned in a format to be used for install_data
    def walk_helper(arg, dirname, files):
        if '.svn' in dirname:
            return
        names = []
        lst, wildcards = arg
        for wc in wildcards:
            wc_name = opj(dirname, wc)
            for f in files:
                filename = opj(dirname, f)

                if not any(bad in filename for bad in badnames):
                    if fnmatch.fnmatch(filename, wc_name)\
                            and not os.path.isdir(filename):
                        names.append(filename)
        if names:
            lst.append((dirname, names))

    file_list = []
    recursive = kw.get('recursive', True)
    if recursive:
        os.path.walk(srcdir, walk_helper, (file_list, wildcards))
    else:
        walk_helper((file_list, wildcards),
                    srcdir,
                    [os.path.basename(f) for f in glob.glob(opj(srcdir, '*'))])
    return file_list
files = find_data_files('sacrud/', '*.*')
print 'files', files

setup(
    name='sacrud',
    version=version.__version__,
    url='http://github.com/uralbash/sacrud/',
    author='Svintsov Dmitry',
    author_email='root@uralbash.ru',

    packages=['sacrud', 'sacrud.pyramid_ext'],
    data_files=files,
    include_package_data=True,
    zip_safe=False,
    test_suite="sacrud.tests",
    license="GPL",
    package_dir={'sacrud': 'sacrud'},
    package_data={
        'sacrud': ['static/style/*.css',
                   'static/js/*.js',
                   'static/js/jquery/*.js',
                   'static/images/*',
                   'templates/*.jinja2', 'templates/forms/*.jinja2',
                   'templates/forms/actions/*.jinja2',
                   'templates/types/*.jinja2',
                   'tests/*.py', ],
        'sacrud.pyramid_ext': ['tests/*.py']
    },
    description='SQLAlchemy CRUD.',
    long_description=open('README.txt').read(),
    install_requires=[
        "sqlalchemy",
        "transaction",
        'zope.sqlalchemy',
        'webtest',
    ],
    classifiers=[
            'Development Status :: 3 - Alpha',
            'Environment :: Console',
            'Environment :: Web Environment',
            'Intended Audience :: Developers',
            'License :: OSI Approved :: GNU General Public License (GPL)',
            'Natural Language :: English',
            'Natural Language :: Russian',
            'Operating System :: OS Independent',
            'Programming Language :: Python',
            'Topic :: Database', ],
)



import os
from sqlalchemy.types import TypeDecorator, VARCHAR


class FileStore(TypeDecorator):
    impl = VARCHAR

    def __init__(self, path='', abspath='', *arg, **kw):
        TypeDecorator.__init__(self, *arg, **kw)
        self.path = path
        self.abspath = abspath

    def process_bind_param(self, value, dialect):
        if value is not None:
            if hasattr(value, 'filename'):
                return os.path.join(self.path, value.filename)
            if 'http://' in value or 'https://' in value:
                return value
            return os.path.join(self.path, value)

    def process_result_value(self, value, dialect):
        if value is not None:
            value = value
        return value
